#!/usr/bin/env python
# coding: utf8
# (c) 2014 Dominic Springer
# This file is licensed under the 2-Clause BSD License (see HARP_License.txt)

import re
import numpy as np
import xml.etree.ElementTree as ET 
import os

from Imports_Basic import *

#====================================================================================
# REGEX MINIMAL HOWTO
#====================================================================================
# re.search and re.sub accept a) a regex string or b) a precompiled version
# if the regex is complicated, compile it first (e.g. with re.MULTILINE|re.DOTALL)
# see replace_Blocks() for an example
#
# NOTES:
# re.MULTILINE: allow "^" to match a newline INSIDE the string (not standard)
# re.DOTALL: the dot "." now also matches a newline (not standard)

#====================================================================================
# NOTES
#====================================================================================
# search() returns SRE_Match object, Python doc refers to this as "MatchObject"

#==========================================================================================================================
#=============================================== REGEX HELPERS ==========================================================
#==========================================================================================================================

#==========================================
def createRegexOR(tokens = (".", "!", "?")):
#==========================================
    # interesting: "map" makes sure that special chars are escaped!
    # Example: \.|\!|\?
    regexPattern = '|'.join(map(re.escape, tokens))
    return regexPattern

#==========================================
def splitWithRegex(String, RegexPattern):
#==========================================
    #will return a simple list
    return re.split(RegexPattern, String, maxsplit=0, flags=re.MULTILINE|re.DOTALL) #maxsplit: once big junk at end, after max splits
    

#==========================================================================================================================
#=============================================== REPLACE IN FILE ==========================================================
#==========================================================================================================================

#NOTE: fileinput redirects stdout, so print will actually "rewrite" to file
#NOTE: *.bak fikes

#==========================================
def replaceInFile_Simple(FN, ToReplace, Replacement):
#==========================================
    print "Replacing in %s: \nOLD: \n%s \nNEW: \n%s" % (FN, ToReplace, Replacement)
    if 0: #example
        ToReplace = \
"""        double getAscent();
        double getDescent();"""
        
        Replacement = \
"""        double getAscent()
        double getDescent()"""

    
    for line in fileinput.input(FN, inplace=True, backup='.bak'):
        newline = line.replace(ToReplace, Replacement)
        if newline != line:
            sys.stderr.write('Replacements. Line now reads: ' + newline)
            
        print newline,
        
#==========================================
def replaceInFile_Simple_Multiline(FN, ToReplace, Replacement):
#==========================================
    with open(FN, 'r') as f:
      content = f.read()
    with open(FN+ ".bak", 'w') as f:
      f.write(content)   
    
    content = content.replace(ToReplace, Replacement)
    with open(FN, 'w') as f:
      f.write(content)

#NOTE: simply an additional regex, nothing fancy compared to above
#==========================================
def replaceInFile_Regex(FileName, SearchFor, ReplaceWith):
#==========================================
    print "Replacing %s with %s in %s " % (SearchFor, ReplaceWith, FileName)
    import fileinput
    for line in fileinput.input(FileName, inplace=True, backup='.bak'): #WARNING: will delete the orig file!
        line = re.sub(SearchFor, ReplaceWith, line)
        print "%s" % (line),  #Back to file

#==========================================================================================================================
#================================================== LOOKAHEAD =============================================================
#==========================================================================================================================

#negative lookahead:  q(?!u)  <- u is a trivial regex
#example (all lines except those which start with a whitespace):
# ^(?!\s)(.*)

#positive lookahead:  q(?=u)  <- u is a trivial regex
#example (all lines THAT ACTUALLY START with a whitespace):
# ^(?=\s)(.*)

# Example: '(Sec.*?)(?=[A-z])'  # FIXES PROBLEMS WITH GROUP REPLACEMENTS (WHY DOES REPLACEMENT ON GROUPS NOT PROPERLY WORK??)


#==========================================================================================================================
#================================================== SEARCH ================================================================
#==========================================================================================================================

# interesting test: obviously there is a difference between match and search... :)
if 0:
    p = re.compile('[a-z]+')
    mymatch = p.match('::: message message')
    mysearch = p.search('::: message message') ; 
    
    print "mymatch: ", mymatch
    print "mysearch: ", mysearch
    print "mysearch.group(): ", mysearch.group()
    print "mysearch.span(): ", mysearch.span()


#interesting! Non-capturing group! (-> (?:abcabc)

# NOTE: PYTHEX: simply copy over regex, but use r"..." to prevent backslash horror

#==========================================
def replace_Part(Text, Regex = r'(\\\[.*?\\\])', ReplaceWith = ""):
#==========================================    
    #in Pythex: SAME STRING. HOORAY!
    # DO NOT FORGET TO STATE THE "RAW" STRING!
    # OTHERWISE: DEBUG FOR 1HR...
    comp = re.compile(Regex, re.MULTILINE|re.DOTALL)
    Text = comp.sub(ReplaceWith, Text) 
    return Text

# INTERESTING! Simple method for iterating over groups
#==========================================
def regex_getGroups(Text, Regex=r"(.*\.(?:m|py|json))(.*)"): #NOW works multiple lines
#==========================================
    comp = re.compile(Regex, re.MULTILINE|re.DOTALL)
    res_tuple = comp.findall(Text)
    return res_tuple

# untested
#==========================================
def regex_get_Kpts(String):
#==========================================
    res = re.search(r"(\d*)kpts", String)
    return np.int32(res.group(1))

#untested
#==========================================
def regex_get_Filled(String):
#==========================================
    res = re.search(r"_(.*?)Mask", String)
    return str(res.group(1))    

#==========================================
def get_Alias_From_FN(PFN):
#==========================================
    Dir, FN = os.path.split(PFN)
    res = re.search(r"(.*)_(\d*)x(\d*)\.yuv", FN)
    return res.group(1)

#==========================================
def get_DimX_DimY_from_Filename(PFN):
#==========================================
    Dir, FN = os.path.split(PFN)
    res = re.search(r"(\d*)x(\d*)\.yuv", FN)
    return np.int32(res.group(1)), np.int32(res.group(2))

    #-
    #ALSO:
    #get DimX and DimY from header in file
    #         line = FH.readline()      
    #         DimX = int( re.search(r"DimX=([-|\d]*)", line).group(1))
    #         DimY = int( re.search(r"DimY=([-|\d]*)", line).group(1))
    #-
        
#==========================================
def get_CTU_CallCnt_from_Filename(FN):
#==========================================
    #example: get_Special_Predictor_CTU059_CallCnt045.hdt
    res = re.search(r"CTU(\d*)_CallCnt(\d*)", FN)
    return (np.int32(res.group(1)), np.int32(res.group(2)))

#==========================================
def get_POCNum_from_FN(FN):
#==========================================
    return int(re.search(r".*POC_(\d*)\D", FN).group(1))

#==========================================
def get_QP_from_Call(Call):
#==========================================
    return int(re.search(r"--QP=(\d*)", Call).group(1))

#==========================================
def get_NumFr_from_Call(Call):
#==========================================
    return int(re.search(r"-fr\s(\d\d)", Call).group(1))

#==========================================
def fromStartToBracket(Str):
#==========================================
    res = re.search(r"::(.*)\sat", Str)
    return (res.group(1))

#==========================================================================================================================
#================================================== REPLACE ===============================================================
#==========================================================================================================================

#==========================================
def doxygen_replace(Content, ReplaceList):
#==========================================
    for token in ReplaceList:
        printLARGE('REPLACING: "%s"' % token)
        LeftToken = token.split("=")[0] + "="
        print "Searching: ", LeftToken 
        Regex = r'%s.*' % LeftToken
        comp = re.compile(Regex)
        #Text = comp.sub(ReplaceWith, Text) 
        res_tuple = comp.findall(Content)
        print "Found:"
        pprint(res_tuple)
        assert(len(res_tuple) == 1) #error if multiple matches
        print "REPLACING!"
        Content = comp.sub(token, Content) 
    return Content


#==========================================
def replace_COMMENTBLOCK(Content, To=""):
#==========================================
    RegEx = '(\% *\n\% Written by:.*?% \n%-----)'
    comp = re.compile(RegEx, re.MULTILINE|re.DOTALL)
    Content = comp.sub(To, Content) 
    return Content

#==========================================
def rename_Files(Path, replace="ASIFT_H", withthis="GT_H"):
#==========================================
    import glob
    FNs = glob.glob(Path + "/*%s*" % replace) #FNs will have SearchDir as Prefix
    for FN in FNs:
        Dir, BN, Suffix = split_into_Path_BN_Suffix(FN)
        PureFN = BN + Suffix
        comp = re.compile(replace)
        NewPureFN = comp.sub(withthis, PureFN) 
        #print NewPureFN
        NewFN = Dir + "/" + NewPureFN
        print NewFN
        moveFile(FN, NewFN)

#==========================================
def replace_brackets(Line, ReplaceWith = ""):
#==========================================    
    comp = re.compile('(<.*?>)')
    content = comp.sub(ReplaceWith, Line) 
    return content

#==========================================
def replace_quotes(Line, ReplaceWith = ""):
#==========================================    
    #Line = Line.replace("'", ReplaceWith)
    #return Line
    comp = re.compile(r"(')")
    content = comp.sub(ReplaceWith, Line) 
    return content

#==========================================
def replace_BLOCKs(FN, isDry):
#==========================================
    # Zeilenanfang, beliebig viele Whitespaces
    # Dann //-  mit Leerzeichen danach (sonst Verwechslungsgefahr)
    # Beliebig viele Zeichen bis zum nächsten //-, allerdings non-greedy ("?")
    
    content = open(FN).read()
    isCPP = True if getSuffix(FN) in (".c", ".cpp", ".h") else False
    
    #------
    if isCPP: 
        comp = re.compile('^\s*//>>>\s.*?//<<<\s', re.MULTILINE|re.DOTALL)
    else:     
        comp = re.compile('^\s*#>>>\s.*?#<<<\s', re.MULTILINE|re.DOTALL)
    #------
    
    found_tuple = comp.findall(content)
    if len(found_tuple) > 0:
        printLARGE("replace_BLOCKs(): File=" + pureFN(FN))
        for hit in found_tuple:
            print hit
        NewContent = comp.sub("", content)    
        open(FN, 'w').write(NewContent)   

#==========================================
def replace_TODOs(FN, isDry):
#==========================================
    content = open(FN).read()
    isCPP = True if getSuffix(FN) in (".c", ".cpp", ".h") else False
    
    #------
    if isCPP: 
        comp = re.compile('^\s*//\s*(?:TODO|TBD).*?\r?$', re.MULTILINE|re.DOTALL)
    else:     
        comp = re.compile('^\s*#\s*(?:TODO|TBD).*?\r?$', re.MULTILINE|re.DOTALL)
    #------
    
    found_tuple = comp.findall(content)
    if len(found_tuple) > 0:
        printLARGE("replace_TODOs(): File=" + pureFN(FN))
        for hit in found_tuple:
            print hit
        NewContent = comp.sub("", content)    
        if not isDry:
            open(FN, 'w').write(NewContent)  
            


#==========================================
def replace_AFFINEs(content, isCpp):
#==========================================

    # Macros
    comp = re.compile('AFFINE')
    content = comp.sub("SPECIAL", content) 
    
    # Anything else
    comp = re.compile('AfFinE', re.IGNORECASE)
    content = comp.sub("Special", content)    
    return content   


    
# FOR PREPARE_RELEASE SCRIPT:

#==========================================================================================================================
#================================================== XML ===================================================================
#==========================================================================================================================


#==========================================
def find_XML_Element(Parent, EclKey):
#==========================================
    for Element in Parent:
        for (AttrName, AttrValue) in Element.items():
            # "Attribute Key" and "Attribute Value": XML terminology
            # "Key" and "Value": names that Eclipse defines (we have some collision here)
            if AttrName == "key" and AttrValue == EclKey:
                return Element
    assert(0), "Element not found"

#==========================================
def print_XML_Tree_Depth1(RootNode):
#==========================================    
    print RootNode.tag, RootNode.attrib
    for child_of_root in RootNode:
        print child_of_root.tag, child_of_root.attrib   
   
#==========================================
def print_XML_Element(Element):
#========================================== 
    print "Tag: ". Element.tag
    print "Attribute: ", Element.attrib
    print "Value of attribute: ", Element.attrib["value"]
    

#==========================================================================================================================
#================================================== VARIOUS ===============================================================
#==========================================================================================================================

